"""
Author: Wenyu Ouyang
Date: 2025-01-07 09:05:05
LastEditTime: 2025-01-07 09:06:26
LastEditors: Wenyu Ouyang
Description: Convention for table names and columns
FilePath: \hydrodatasource\hydrodatasource\configs\table_name.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
# reservoir timeseries table columns
RSVR_TS_TABLE_COLS = [
    "STCD",
    "TM",
    "RZ",
    "INQ",
    "W",
    "BLRZ",
    "OTQ",
    "RWCHRCD",
    "RWPTN",
    "INQDR",
    "MSQMT",
]
